package edu.cs2335.SWATLP.client.gui;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.PrintWriter;
import java.net.Socket;

import javax.swing.JOptionPane;
import javax.swing.SwingUtilities;

import edu.cs2335.SWATLP.client.clientdata.PlayerData;

/**
 * ClientConnection - the connection to the network.
 */
public class ClientConnection implements Runnable {
    /**
     * Reference to the client main class.
     */
    private MainClass main;

    /**
     * The connection.
     */
    private Socket clientsocket;

    /**
     * This player's name.
     */
    private String playerName;

    /**
     * PrintWriter to send strings over the network.
     */
    private PrintWriter output;

    /**
     * BufferedReader to read messages from the network.
     */
    private BufferedReader input;

    /**
     * Input stream.
     */
    private InputStreamReader instream;

    /**
     * Input stream to read objects from server.
     */
    private ObjectInputStream objectInput;

    /**
     * Output stream to send objects to the server.
     */
    private ObjectOutputStream objectOutput;

    /**
     * Controls whether execution of the thread continues.
     */
    private boolean running;

    /**
     * Game test
     */
    private boolean gamestate;

    /**
     * Constructor that sets up the thread.
     * 
     * @param mainclass
     *            The control class
     * @param hostname
     *            host to connect to
     * @param port
     *            the port to connect to
     * @throws IOException
     *             thrown if errors
     */
    public ClientConnection(MainClass mainclass, String hostname, String port)
            throws IOException {
        running = true;
        main = mainclass;
        clientsocket = new Socket(hostname, Integer.parseInt(port));
        output = new PrintWriter(clientsocket.getOutputStream());
        instream = new InputStreamReader(clientsocket.getInputStream());
        input = new BufferedReader(instream);
        gamestate = false;
        sendMessage("OK");
    }

    /**
     * Main loop. Receives messages from the server.
     */
    public void run() {
        String messagefromserver = null;
        String[] message = null;
        try {
            while (running) {
                messagefromserver = input.readLine();
                if (messagefromserver == null) {
                    running = false;
                    return;
                }
                message = messagefromserver.split("~");
                parseMessage(message);
            }
        } catch (IOException ioe) {
            running = false;
            final IOException e = ioe;
            SwingUtilities.invokeLater(new Runnable() {
                public void run() {
                    e.printStackTrace();
                    main.displayError("SF", "connecterror");
                }
            });
        }
    }

    /**
     * Parses the message read by the server and calls an appropriate method
     * 
     * @param text
     *            an array containing the tokens of the message
     */
    private void parseMessage(String[] text) {
        // server sent: someone is chatting
        if (text[0].equals("MAINMSG")) {
            main.displayMainChatMessage(text[1] + text[2]);

            // server sent: you created a new character
        } else if (text[0].equals("RS")) {
            displayMessage(text);

            // server sent: some player data
        } else if (text[0].equals("PLAYERDATA")) {
            main.playerData(text);

            // This is the player location stuff
        } else if (text[0].equals("PLAYERLOC")) {
            main.playerLoc(text);

            // server sent: shipyard data
        } else if (text[0].equals("SHIPYARDDATA")) {
            main.getGui().setupShipyard(text);

            // server sent: the planet data
        } else if (text[0].equals("PLANETDATA")) {
            main.getGui().setPlanetData(text[1], text[2], text[3], text[4],
                                        Integer.parseInt(text[5]),
                                        Integer.parseInt(text[6]));
            if (text[7].equals("t")) {
                sendMessage("APPROACHDATA~" + text[1] + "~t");
            } else {
                sendMessage("APPROACHDATA~" + text[1] + "~f");
            }

            // server sent: the port data
        } else if (text[0].equals("PORTDATA")) {
            main.getGui().setPlanetData(text[1], text[2], text[3], text[4],
                                        Integer.parseInt(text[5]),
                                        Integer.parseInt(text[6]));
            main.getGui().setupPort();
            
            // server sent: marketplace data
        } else if (text[0].equals("MARKETDATA")) {
            main.getGui().setupMarketplace(text);

            // server sent: all the approach data
        } else if (text[0].equals("APPROACHDATA")) {
            main.getGui().setupApproach(text);

        } else if (text[0].equals("MARKETTRANS")) {
            main.getGui().updateMarketplace(text);

            // server sent: bank data
        } else if (text[0].equals("BANKDATA")) {
            main.getGui().displayBankWindow(text); 

            // server sent: range chart data
        } else if (text[0].equals("RANGECHARTDATA")) {
            main.getGui().setupChart(text);
            
            // server sent: range chart data
        } else if (text[0].equals("SHIPYARDTRANS")) {
            main.displayError(text[1], text[2]);

        } else if (text[0].equals("WAREHOUSEDATA")) {
            main.getGui().setupWarehouse(text);
            // server sent: an error
        } else if (text[0].equals("SCHOOLDATA")) {
            main.getGui().setupSchool(text);
        } else if (text[0].equals("X")) {
            displayErrorMessage(text);

            // server sent: the server's shutting down
        } else if (text[0].equals("SF")) {
            shutdownClient();
            displayShutdownMessage();
        } else {
            displayErrorMessage(text);
        }
    }

    /**
     * Displays a notice to the user that something occurred (nonerror).
     * 
     * @param message
     *            the message to display
     */
    private void displayMessage(String[] message) {
        String sentmessage = null;
        if (message[0].equals("RS")) {
            sentmessage = "Character has been created.";
            main.getGui().setUser(message[1], message[2]);
            PlayerData data = main.getGui().getPlayerData();
            data.setName(message[1]);
            data.setPass(message[2]);
            data.setPlanet(message[3]);
            data.resetPlanetsList();
            main.getGui().setPlayerData(data);
            
        }
        final String msg = sentmessage;
        SwingUtilities.invokeLater(new Runnable() {
            public void run() {
                main.displayMessage(msg);
            }
        });
    }

    /**
     * Displays a message to the user when an error occurs.
     * 
     * @param message
     *            a message describing the error
     */
    private void displayErrorMessage(String[] message) {
        String errormessage = "";
        String errorcode = "";
        if (message.length > 1) {
            errorcode = message[1];
        }
        if (message.length > 2) {
            errormessage = message[2];
        }
        final String titl = errorcode;
        final String msg = errormessage;
        SwingUtilities.invokeLater(new Runnable() {
            public void run() {
                main.displayError(titl, msg);
            }
        });
    }

    /**
     * Shuts down this thread and closes connections.
     */
    public synchronized void shutdownClient() {
        try {
            if (gamestate) {
                sendMessage("LEAVEGAME");
            }
            sendMessage("QUIT");
            running = false;
            main.setNewconnection(true);
            input.close();
            output.close();
            instream.close();
            clientsocket.close();
        } catch (IOException e) {
            e.printStackTrace();
            running = false;
            return;
        }
    }

    /**
     * Display the shutdown message.
     */
    private void displayShutdownMessage() {
        SwingUtilities.invokeLater(new Runnable() {
            public void run() {
                String title = "Server Failure";
                String displaymessage = "The server shutdown. You have "
                        + "been logged off.";
                if (JOptionPane.showConfirmDialog(null, displaymessage, title,
                        JOptionPane.DEFAULT_OPTION,
                        JOptionPane.INFORMATION_MESSAGE) 
                            == JOptionPane.YES_OPTION) {
                    System.exit(0);
                }
            }
        });
    }

    /**
     * use this to send a message to the server
     * 
     * @param msg
     *            the message to send
     */
    public final void sendMessage(String msg) {
        output.println(msg);
        output.flush();
        System.out.println("Client Sent: " + msg);
    }

    /**
     * @return Returns the username.
     */
    public String getPlayerName() {
        return playerName;
    }

    /**
     * @param playerName
     *            The username to set.
     */
    public void setPlayerName(String playerName) {
        this.playerName = playerName;
    }

    /**
     * @return Returns the client's socket.
     */
    public Socket getClientSocket() {
        return clientsocket;
    }
    /**
     * @return Returns the clientsocket.
     */
    public Socket getClientsocket() {
        return clientsocket;
    }
    /**
     * @param clientsocket The clientsocket to set.
     */
    public void setClientsocket(Socket clientsocket) {
        this.clientsocket = clientsocket;
    }
    /**
     * @return Returns the gamestate.
     */
    public boolean isGamestate() {
        return gamestate;
    }
    /**
     * @param gamestate The gamestate to set.
     */
    public void setGamestate(boolean gamestate) {
        this.gamestate = gamestate;
    }
    /**
     * @return Returns the input.
     */
    public BufferedReader getInput() {
        return input;
    }
    /**
     * @param input The input to set.
     */
    public void setInput(BufferedReader input) {
        this.input = input;
    }
    /**
     * @return Returns the instream.
     */
    public InputStreamReader getInstream() {
        return instream;
    }
    /**
     * @param instream The instream to set.
     */
    public void setInstream(InputStreamReader instream) {
        this.instream = instream;
    }
    /**
     * @return Returns the main.
     */
    public MainClass getMain() {
        return main;
    }
    /**
     * @param main The main to set.
     */
    public void setMain(MainClass main) {
        this.main = main;
    }
    /**
     * @return Returns the objectInput.
     */
    public ObjectInputStream getObjectInput() {
        return objectInput;
    }
    /**
     * @param objectInput The objectInput to set.
     */
    public void setObjectInput(ObjectInputStream objectInput) {
        this.objectInput = objectInput;
    }
    /**
     * @return Returns the objectOutput.
     */
    public ObjectOutputStream getObjectOutput() {
        return objectOutput;
    }
    /**
     * @param objectOutput The objectOutput to set.
     */
    public void setObjectOutput(ObjectOutputStream objectOutput) {
        this.objectOutput = objectOutput;
    }
    /**
     * @return Returns the output.
     */
    public PrintWriter getOutput() {
        return output;
    }
    /**
     * @param output The output to set.
     */
    public void setOutput(PrintWriter output) {
        this.output = output;
    }
    /**
     * @return Returns the running.
     */
    public boolean isRunning() {
        return running;
    }
    /**
     * @param running The running to set.
     */
    public void setRunning(boolean running) {
        this.running = running;
    }
}