package edu.cs2335.SWATLP.server.main;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketException;
import java.util.ArrayList;
import java.util.List;

/**
 * SWATLP! - CS2335 Lab8
 * 
 * @version 1.0 <p/>Server - The main server class. Controls server startup and
 *          accepts client connections. Performs different operations involving
 *          multiple clients.
 */
public class Server implements Runnable {
    /**
     * Main server socket, only one created.
     */
    private ServerSocket mainsocket;

    /**
     * Client socket, one created for each connected client.
     */
    private Socket clientsocket;

    /**
     * Buffer to read input from client sockets.
     */
    private BufferedReader input;

    /**
     * A list of all client connections.
     */
    private List connections;

    /**
     * Reference to the database of user information.
     */
    private PlayerDatabase database;

    /**
     * The universe!
     */
    private Universe gameworld;
    /** the port to connect to */
    private int portnum;

    /**
     * Server Constructor Binds server to port given and attempts to read the
     * database from disk. If no database exists it creates a new one.
     * 
     */
    public void setupServer() {
        connections = new ArrayList();

        // start a new server, load the user database
        try {
            mainsocket = new ServerSocket(portnum);
            FileInputStream fileinput = new FileInputStream("userdatabase.db");
            ObjectInputStream objinput = new ObjectInputStream(fileinput);
            database = (PlayerDatabase) (objinput.readObject());
            objinput.close();
            fileinput.close();
        } catch (FileNotFoundException fe) {
            database = new PlayerDatabase();
            gameworld = new Universe();
            return;
        } catch (ClassNotFoundException ce) {
            ce.printStackTrace();
        } catch (IOException ioe) {
            ioe.printStackTrace();
        }

        // load the universe
        gameworld = database.getGameworld();
    }

    /**
     * Begins accepting connections. When a connection is received a new thread
     * is created and the loop repeats to accept more. The only commands that
     * will be accepted here are EXIT for shutting down the server, NEWUSER for
     * someone registering a new username/password, and LOGIN for someone
     * logging in using an existing account. Anything else is ignored.
     */
    public void startServer() {
        System.out.println("SERVER: starting server");
        try {
            while (true) {
                clientsocket = mainsocket.accept();
                input = new BufferedReader(new InputStreamReader(clientsocket
                        .getInputStream()));
                String readinput = input.readLine();
                if (readinput == null) {
                    continue;
                }
                String[] firstmessage = readinput.split("~");
                if (firstmessage[0].equals("EXIT")) {
                    shutdownServer();
                    return;
                } else {
                    ServerConnection newconnection = new ServerConnection(
                            clientsocket, this, database);
                    System.out.println("SERVER: new connection");
                    System.out.flush();
                    connections.add(newconnection);
                    Thread connectionthread = new Thread(newconnection);
                    connectionthread.start();
                }
            }
        } catch (SocketException e) {
            System.out.println("client connection reset");
        } catch (IOException ioe) {
            ioe.printStackTrace();
        }
    }

    /**
     * When the server gets a shutdown message the database is saved to the
     * disk, all connected clients are notified of the shutdown, and the server
     * exits.
     */
    public void shutdownServer() {
        ServerConnection cn;
        for (int i = 0; i < connections.size(); i++) {
            cn = (ServerConnection) (connections.get(i));
            cn.sendMessage("SF~SHUTDOWN");
        }
        saveGame();
        try {
        mainsocket.close();
        } catch (Exception e) {
            return;
        }
    }

    /**
     * @param messagetosend
     *            a message sends a message to all connected client sockets
     */
    public void sendMessage(String messagetosend) {
        ServerConnection conn;
        System.out.println(connections.size());
        for (int i = 0; i < connections.size(); i++) {
            conn = (ServerConnection) (connections.get(i));
            System.out.println("user:" + conn.getUsername());
            conn.sendMessage(messagetosend);
        }
    }
    
    /**
     * 
     */
    public void saveGame() {
        if (database != null) {
        database.setGameworld(gameworld);
        }
        try {
            FileOutputStream fileoutput = new FileOutputStream(
                    "userdatabase.db");
            ObjectOutputStream objoutput = new ObjectOutputStream(fileoutput);
            objoutput.writeObject(database);
            objoutput.flush();
            objoutput.close();
            fileoutput.close();
        } catch (IOException ioe) {
            ioe.printStackTrace();
        }
    }

    /**
     * when a client thread detects that a client has closed their connection
     * this method is called to remove them from the connection list
     * 
     * @param username
     *            the user to disconnect
     */
    public void disconnectUser(String username) {
        ServerConnection conn = null;
        for (int i = 0; i < connections.size(); i++) {
            conn = (ServerConnection) (connections.get(i));
            if ((conn.getUsername()).equals(username)) {
                connections.remove(i);
            }
        }
    }

    /**
     * Gets the user server connection from the user list
     * 
     * @param usernm
     *            the user name
     * @return there server connection
     */
    public ServerConnection getPlayerConnection(String usernm) {
        ServerConnection conn;
        for (int i = 0; i < connections.size(); i++) {
            conn = (ServerConnection) (connections.get(i));
            if ((conn.getUsername()).equals(usernm)) {
                return conn;
            }
        }
        return null;
    }

    /**
     * Starts the server from the command line. Prints a usage error if port
     * number is entered incorrectly.
     */
    public void run() {
        setupServer();
        startServer();
    }

    /**
     * @return Returns the gameworld.
     */
    public Universe getGameworld() {
        return gameworld;
    }

    /**
     * @param gameworld
     *            The gameworld to set.
     */
    public void setGameworld(Universe gameworld) {
        this.gameworld = gameworld;
    }
    /**
     * @param portnum The portnum to set.
     */
    public void setPortnum(int portnum) {
        this.portnum = portnum;
    }
    /**
     * @return Returns the clientsocket.
     */
    public Socket getClientsocket() {
        return clientsocket;
    }
    /**
     * @param clientsocket The clientsocket to set.
     */
    public void setClientsocket(Socket clientsocket) {
        this.clientsocket = clientsocket;
    }
    /**
     * @return Returns the connections.
     */
    public List getConnections() {
        return connections;
    }
    /**
     * @param connections The connections to set.
     */
    public void setConnections(List connections) {
        this.connections = connections;
    }
    /**
     * @return Returns the database.
     */
    public PlayerDatabase getDatabase() {
        return database;
    }
    /**
     * @param database The database to set.
     */
    public void setDatabase(PlayerDatabase database) {
        this.database = database;
    }
    /**
     * @return Returns the input.
     */
    public BufferedReader getInput() {
        return input;
    }
    /**
     * @param input The input to set.
     */
    public void setInput(BufferedReader input) {
        this.input = input;
    }
    /**
     * @return Returns the mainsocket.
     */
    public ServerSocket getMainsocket() {
        return mainsocket;
    }
    /**
     * @param mainsocket The mainsocket to set.
     */
    public void setMainsocket(ServerSocket mainsocket) {
        this.mainsocket = mainsocket;
    }
    /**
     * @return Returns the portnum.
     */
    public int getPortnum() {
        return portnum;
    }
}